#define RASTER_W		256
#define RASTER_H		240

#define SAMPLE_RATE		20000 //picked to divide the 80 MHz timer source clock evenly

#define SOUND_VOLUME	100

enum {
  EMU_FOTON,
  EMU_TIA88,
  EMU_TIA91
};

enum {
  GAME_PITON = 0,
  GAME_KLAD,
  GAME_TETRIS,
  GAME_KONEK,
  GAME_KOROLEVA,
  GAME_SOS,
  GAME_GORODKI,
  GAME_BILYARD,
  GAME_KOT_RYBOLOV,
  GAMES_ALL
};

int emu_core;
int emu_loops;

int pad_left;
int pad_right;
int pad_up;
int pad_down;
int pad_a;
int pad_b;
int pad_coin;

#define AUDIO_QUEUE_LEN		(SAMPLE_RATE/50*8)

unsigned char audio_queue[AUDIO_QUEUE_LEN];

volatile int audio_queue_wr;
volatile int audio_queue_rd;
volatile int audio_queue_len;

void audio_queue_init(void)
{
  memset(audio_queue, 0, sizeof(audio_queue));

  audio_queue_wr = 0;
  audio_queue_rd = 0;
  audio_queue_len = 0;
}

inline void audio_queue_add(unsigned char n)
{
  if (audio_queue_len < AUDIO_QUEUE_LEN)
  {
    audio_queue[audio_queue_wr++] = n;

    if (audio_queue_wr >= AUDIO_QUEUE_LEN) audio_queue_wr = 0;

    ++audio_queue_len;
  }
}

inline unsigned char audio_queue_get(void)
{
  unsigned char n = 0;

  if (audio_queue_len > 0)
  {
    n = audio_queue[audio_queue_rd++];

    if (audio_queue_rd >= AUDIO_QUEUE_LEN) audio_queue_rd = 0;

    --audio_queue_len;
  }

  return n;
}

bool audio_queue_is_empty(void)
{
  return audio_queue_len == 0 ? true : false;
}

bool audio_queue_is_half_empty(void)
{
  return audio_queue_len < (AUDIO_QUEUE_LEN / 2) ? true : false;
}

#include "CPU.h"
#include "IOList.h"
#include "EmuFoton.h"
#include "SoundTIA.h"
#include "EmuTIA88.h"
#include "EmuTIA91.h"


void set_emu_defaults(void)
{
  pad_left = PAD_LEFT;
  pad_right = PAD_RIGHT;
  pad_up = PAD_UP;
  pad_down = PAD_DOWN;
  pad_a = PAD_A;
  pad_b = PAD_B;
  pad_coin = PAD_COIN;

  emu_loops = 2;	//fill the audio buffer beforehand
}

void set_emu_foton(void)
{
  emu_core = EMU_FOTON;

  set_emu_defaults();

  EmuFoton::init();
}

void set_emu_tia88(bool analog)
{
  emu_core = EMU_TIA88;

  set_emu_defaults();

  EmuTIA88::init(analog);
}

void set_emu_tia91(void)
{
  emu_core = EMU_TIA91;

  set_emu_defaults();

  EmuTIA91::init();
}

void load_piton(void)
{
  set_emu_foton();

  EmuFoton::rom_load("foton_piton.bin", 0x0000);

  emu_loops = 70;
}

void load_klad(void)
{
  set_emu_foton();

  EmuFoton::rom_load("foton_klad.bin", 0x0000);
}

void load_tetris(void)
{
  set_emu_foton();

  pad_a = PAD_UP;
  pad_b = PAD_DOWN;

  EmuFoton::rom_load("foton_tetris.bin", 0x0000);
}

void load_konek(void)
{
  set_emu_tia88(false);

  EmuTIA88::pmem_load("konek_g1.bin", 0x0000);
  EmuTIA88::pmem_load("konek_g2.bin", 0x2000);
  EmuTIA88::pmem_load("konek_g3.bin", 0x4000);
  EmuTIA88::pmem_load("konek_g4.bin", 0x6000);
  EmuTIA88::pmem_load("konek_g5.bin", 0x8000);
  EmuTIA88::pmem_load("konek_g7.bin", 0xc000);

  EmuTIA88::vrom_load("konek_a2.bin", 0x0000);
  EmuTIA88::vrom_load("konek_a3.bin", 0x2000);
  EmuTIA88::vrom_load("konek_a5.bin", 0x4000);
  EmuTIA88::vrom_load("konek_a6.bin", 0x6000);

  emu_loops = 500;
}

void load_bilyard(void)
{
  set_emu_tia88(false);

  EmuTIA88::pmem_load("bilyard_g1.bin", 0x0000);
  EmuTIA88::pmem_load("bilyard_g2.bin", 0x2000);
  EmuTIA88::pmem_load("bilyard_g7.bin", 0xc000);

  EmuTIA88::vrom_load("bilyard_a2.bin", 0x0000);
  EmuTIA88::vrom_load("bilyard_a3.bin", 0x2000);
  EmuTIA88::vrom_load("bilyard_a5.bin", 0x4000);
  EmuTIA88::vrom_load("bilyard_a6.bin", 0x6000);

  emu_loops = 250;
}

void load_gorodki(void)
{
  set_emu_tia88(true);

  EmuTIA88::pmem_load("gorodki_g1.bin", 0x0000);
  EmuTIA88::pmem_load("gorodki_g2.bin", 0x2000);
  EmuTIA88::pmem_load("gorodki_g3.bin", 0x4000);
  EmuTIA88::pmem_load("gorodki_g4.bin", 0x6000);

  EmuTIA88::vrom_load("gorodki_a2.bin", 0x0000);
  EmuTIA88::vrom_load("gorodki_a3.bin", 0x2000);
  EmuTIA88::vrom_load("gorodki_a5.bin", 0x4000);
  EmuTIA88::vrom_load("gorodki_a6.bin", 0x6000);

  emu_loops = 400;
}

void load_koroleva(void)
{
  set_emu_tia88(false);

  EmuTIA88::pmem_load("koroleva_g1.bin", 0x0000);
  EmuTIA88::pmem_load("koroleva_g2.bin", 0x2000);
  EmuTIA88::pmem_load("koroleva_g3.bin", 0x4000);
  EmuTIA88::pmem_load("koroleva_g4.bin", 0x6000);
  EmuTIA88::pmem_load("koroleva_g5.bin", 0x8000);
  EmuTIA88::pmem_load("koroleva_g7.bin", 0xc000);

  EmuTIA88::vrom_load("koroleva_a2.bin", 0x0000);
  EmuTIA88::vrom_load("koroleva_a3.bin", 0x2000);
  EmuTIA88::vrom_load("koroleva_a5.bin", 0x4000);
  EmuTIA88::vrom_load("koroleva_a6.bin", 0x6000);

  emu_loops = 250;
}

void load_sos(void)
{
  set_emu_tia88(false);

  EmuTIA88::pmem_load("sos_g1.bin", 0x0000);
  EmuTIA88::pmem_load("sos_g2.bin", 0x2000);
  EmuTIA88::pmem_load("sos_g3.bin", 0x4000);
  EmuTIA88::pmem_load("sos_g4.bin", 0x6000);
  EmuTIA88::pmem_load("sos_g5.bin", 0x8000);
  EmuTIA88::pmem_load("sos_g7.bin", 0xc000);

  EmuTIA88::vrom_load("sos_a2.bin", 0x0000);
  EmuTIA88::vrom_load("sos_a3.bin", 0x2000);
  EmuTIA88::vrom_load("sos_a5.bin", 0x4000);
  EmuTIA88::vrom_load("sos_a6.bin", 0x6000);

  emu_loops = 100;
}

void load_kotrybol(void)
{
  set_emu_tia91();

  EmuTIA91::srom_load("kotrybol_850.bin", 0x0000);
  EmuTIA91::srom_load("kotrybol_851.bin", 0x2000);
  EmuTIA91::srom_load("kotrybol_852.bin", 0x4000);
  EmuTIA91::srom_load("kotrybol_853.bin", 0x6000);

  EmuTIA91::pmem_load("kotrybol_854.bin", 0x0000);
  EmuTIA91::pmem_load("kotrybol_855.bin", 0x2000);
  EmuTIA91::pmem_load("kotrybol_856.bin", 0x4000);

  EmuTIA91::crom_load("kotrybol_846.bin", 0x0000);
  EmuTIA91::crom_load("kotrybol_847.bin", 0x2000);
  EmuTIA91::crom_load("kotrybol_848.bin", 0x4000);
  EmuTIA91::crom_load("kotrybol_849.bin", 0x6000);

  emu_loops = 100;
}

void load_game(int id)
{
  switch (id)
  {
    case GAME_PITON:	     load_piton(); break;
    case GAME_KLAD:        load_klad(); break;
    case GAME_TETRIS:      load_tetris(); break;
    case GAME_KONEK:       load_konek(); break;
    case GAME_KOROLEVA:    load_koroleva(); break;
    case GAME_SOS:         load_sos(); break;
    case GAME_GORODKI:     load_gorodki(); break;
    case GAME_BILYARD:	   load_bilyard(); break;
    case GAME_KOT_RYBOLOV: load_kotrybol(); break;
  }

  switch (emu_core)
  {
    case EMU_FOTON: EmuFoton::reset(); break;
    case EMU_TIA88: EmuTIA88::reset(); break;
    case EMU_TIA91: EmuTIA91::reset(); break;
  }
}


int emulate(int input_state, int frames)
{
  int input_remap = PAD_NONE;

  if (input_state & PAD_LEFT) input_remap |= pad_left;
  if (input_state & PAD_RIGHT) input_remap |= pad_right;
  if (input_state & PAD_UP) input_remap |= pad_up;
  if (input_state & PAD_DOWN) input_remap |= pad_down;
  if (input_state & PAD_A) input_remap |= pad_a;
  if (input_state & PAD_B) input_remap |= pad_b;
  if (input_state & PAD_COIN) input_remap |= pad_coin;

  bool audio_underrun = audio_queue_is_half_empty();

  int t = 0;

  if (emu_loops < frames) emu_loops = frames;

  IOList::clear();

  while (emu_loops > 0)
  {
    switch (emu_core)
    {
      case EMU_FOTON:
        {
          t += EmuFoton::update(input_remap, audio_underrun);
        }
        break;

      case EMU_TIA88:
        {
          t += EmuTIA88::update(input_remap, audio_underrun);
        }
        break;

      case EMU_TIA91:
        {
          t += EmuTIA91::update(input_remap, audio_underrun);
        }
        break;
    }

    --emu_loops;

    audio_underrun = false;
  }

  return t;
}


void render_video(unsigned char** raster)
{
  if (raster)
  {
    switch (emu_core)
    {
      case EMU_FOTON:
        {
          EmuFoton::render_video(raster);
        }
        break;

      case EMU_TIA88:
        {
          EmuTIA88::render_video(raster);
        }
        break;

      case EMU_TIA91:
        {
          EmuTIA91::render_video(raster);
        }
        break;
    }
  }
}



void render_audio(int t)
{
  switch (emu_core)
  {
    case EMU_FOTON:
      {
        EmuFoton::render_audio(t);
      }
      break;

    case EMU_TIA88:
      {
        EmuTIA88::render_audio(t);
      }
      break;

    case EMU_TIA91:
      {
        EmuTIA91::render_audio(t);
      }
      break;
  }
}
