namespace SoundTIA
{
#define PIT_FREQ				(15750000 / 9)
#define PIT_TICKS_PER_SAMPLE	(PIT_FREQ / SAMPLE_RATE)

int sound_out_acc;
int sound_out_val;
int sound_out_div;

unsigned char pit_load_mode[4];
unsigned char pit_load_flip[4];
unsigned short int pit_counter[4];
unsigned char pit_count_mode[4];

int sound_freq_acc;
int sound_freq_div;

int sound_audible_channel;

void init(int ch)
{
  sound_audible_channel = ch;

  IOList::clear();
}

void reset(void)
{
  sound_out_acc = 0;
  sound_out_val = 0;
  sound_out_div = 0;

  sound_freq_acc = 0;
  sound_freq_div = 0;

  for (int i = 0; i < 4; ++i)
  {
    pit_load_mode[i] = 0;
    pit_load_flip[i] = 0;
    pit_counter[i] = 0;
    pit_count_mode[i] = 0;
  }
}

void write(int time, unsigned char reg, unsigned char val)
{
  IOList::add(time, reg, val);
}

void render(int t)
{
  int prev_t = 0;

  for (int i = 0; i < IOList::length(); ++i)
  {
    int new_t = IOList::entry[i].time;

    int diff = (new_t - prev_t);

    prev_t = new_t;

    //PIT internal logic here

    int reg = IOList::entry[i].reg;
    int val = IOList::entry[i].val;

    switch (reg)
    {
      case 0xc0:
      case 0xc1:
      case 0xc2:
        {
          reg &= 3;

          if (!pit_load_flip[reg])
          {
            pit_counter[reg] = (pit_counter[reg] & 0xff00) | val;
          }
          else
          {
            pit_counter[reg] = (pit_counter[reg] & 0x00ff) | (val << 8);
          }

          if (pit_load_mode[reg] == 3) pit_load_flip[reg] ^= 1;

          if (reg == sound_audible_channel)
          {
            sound_freq_div = 65536 * PIT_TICKS_PER_SAMPLE * 4 / (pit_counter[sound_audible_channel] + 1);

            if (sound_freq_div > SAMPLE_RATE * PIT_TICKS_PER_SAMPLE / 2) sound_freq_div = 0;	//mute if the frequency is too high
          }
        }
        break;

      case 0xc3:
        {
          int ch = val >> 6;

          pit_count_mode[ch] = (val >> 1) & 7;

          pit_load_mode[ch] = (val >> 4) & 3;

          if (pit_load_mode[ch] != 2) pit_load_flip[ch] = 0; else pit_load_flip[ch] = 1;
        }
        break;
    }

    //tone generator

    int samples = diff / PIT_TICKS_PER_SAMPLE;

    for (int j = 0; j < samples; ++j)
    {
      sound_freq_acc += sound_freq_div;

      if (pit_count_mode[sound_audible_channel] == 3)
      {
        if (sound_freq_acc & 0x10000) sound_out_val = 0; else sound_out_val = SOUND_VOLUME;
      }

      audio_queue_add(sound_out_val);
    }
  }
}
}
